/*
 * Copyright (C) 2017-2023 Dominique Dhoutaut, Thierry Arrabal, Eugen Dedu.
 *
 * This file is part of BitSimulator.
 *
 * BitSimulator is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * BitSimulator is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BitSimulator.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <cassert>
#include "raytracing-routing-agent.h"
#include "world.h"
#include "../scheduler.h"

#define RTF_SIZE 10
#define CTF_SIZE 11
#define FORGET_SIZE 12
#define FORGET2_SIZE 13

//==============================================================================
//
//          RayTracingRoutingAgent  (class)
//
//==============================================================================

mt19937_64 *RayTracingRoutingAgent::backoffrng;

void RayTracingRoutingAgent::initializeAgent() {
  backoffrng = new mt19937_64 (ScenarioParameters::getBackoffRNGSeed());
}

RayTracingRoutingAgent::RayTracingRoutingAgent(Node *_hostNode) : RoutingAgent(_hostNode) {
  datacounter[0] = 0;
  rtfSender[0] = -1;
  packetForwarded[0] = false;
}

RayTracingRoutingAgent::~RayTracingRoutingAgent() {
  if ( aliveAgents == 1 )
    cout << " Ray tracing:: forwarded packets: " << forwardedDataPackets << endl;
}


void RayTracingRoutingAgent::receivePacketFromApplication(PacketPtr packet) {
  packet->setSrcSequenceNumber(hostNode->getNextSrcSequenceNumber());
  packet->disableBackoff = true;
  hostNode->enqueueOutgoingPacket(packet);
}

void RayTracingRoutingAgent::receivePacketFromNetwork(PacketPtr packet) {
  // source must not resend the packet
  if ( packet->srcId == hostNode->getId())
    return;

  /*static*/ simulationTime_t tp = Node::getPulseDuration();
  /*static*/ simulationTime_t ts = tp * packet->beta;
  int id = packet->flowSequenceNumber;  // TODO make id unique no matter the src/flow
  assert (id == 0);

  if (packet->type == PacketType::RTF) {
    if (packetForwarded[id]
        && (rtfSender[id] == -1 || rtfSender[id] == packet->srcId)) {
      // send CTF right now
      PacketPtr ctf (new Packet (PacketType::CTF, CTF_SIZE, hostNode->getId(), packet->srcId, -1, -1, id));  // use dstId field to specify destination
      ctf->disableBackoff = true;
      hostNode->enqueueOutgoingPacket (ctf);
      //cout << "eugen ctf  sent by node " << hostNode->getId()  << " at " << Scheduler::now() << endl;
      // memorise RTF sender for next RTF, if any
      rtfSender[id] = packet->srcId;
    } else {
      // discard it
    }
    return;
  }

  if (packet->type == PacketType::CTF) {
    // ignore CTF packets received BEFORE two data packets, as they concern other regions
    if (datacounter[id] < 2)
      return;

    ctfReceived[id] = true;
    // node (like C) having sent its RTF sends the packet
    if (packet->dstId == hostNode->getId()) {
      // send saved packet at right time
      assert (timesecond[id] != -1);
      assert (rtfSent[id]);
      // now + x = timesecond[id] + Ts - Tp + k*Ts, what is the smallest x?
      // (now - timesecond[id] + Tp) + x = k'*Ts
      // x = Ts - (now - timesecond[id] + Tp)%Ts
      // if (x==Ts) x=0;  // works without it too, but adds a Ts delay
      simulationTime_t backoff = ts - (Scheduler::now() - timesecond[id] + tp)%ts;  // să șterg +tp?   off by one, vezi log: 15903212 și 2202103112 (032 vs 031)
      if (backoff == ts)
        backoff = 0;
      backoff += ts;  // node has just received the CTF, and minIntervalBetweenReceiveAndSend (e.g. of 99899) must not artificially increase our backoff
      auto sendEvent = new RayTracingSendEvent (Scheduler::now() + backoff, this, packetSave[id]);
      Scheduler::getScheduler().schedule (sendEvent);
    } else
      // it is normal to receive a CTF after 1 data packet for ex., so unschedule only if exactly 2 copies received
      // implemented above: if (timefirst[id] != -1 && timesecond[id] != -1 && unscheduled[id] == false)
      if (unscheduled[id] == false)
        unscheduled[id] = true;  // unschedule, i.e. make it a noop
    return;
  }

  if (packet->type == PacketType::FORGET || packet->type == PacketType::FORGET2) {
    if (datacounter[id] >= 2)  // like in CTF reception
      ctfReceived[id] = false;
    // if (datacounter[id] == 0) datacounter[id]=1;  // for testing shadowing
//if (datacounter[id] <= 0) cout << "EUGEN node " << hostNode->getId() << endl;
    assert (datacounter[id] > 0);
    datacounter[id] --;
    rtfSent[id] = false;
    unscheduled[id] = false;  // ignore any previous unschedules

    if (packet->type == PacketType::FORGET
        && rtfSender[id] != -1 && rtfSender[id] == packet->transmitterId) {  // I sent a CTF to B, so I am node A
      // send FORGET2 packet, no need of delay because the time slot does not matter
      PacketPtr forget2 (new Packet (PacketType::FORGET2, FORGET2_SIZE, hostNode->getId(), -1, -1, -1, id));
      forget2->disableBackoff = true;
      hostNode->enqueueOutgoingPacket (forget2);
      //cout << "eugen forget2 sent by node " << hostNode->getId() << " at " << Scheduler::now() << endl;
      packetForwarded[id] = false;
      rtfSender[id] = -1;

      // send data again
      // copied from above
      simulationTime_t backoff = ts - (Scheduler::now() - timesecond[id] + tp)%ts;  // să șterg +tp?   off by one, vezi log: 15903212 și 2202103112 (032 vs 031)
      if (backoff == ts)
        backoff = 0;
      backoff += 1000*ts;  // node has just started to send a FORGET2 packet, and BS cannot send two packets concurrently
      auto sendEvent = new RayTracingSendEvent (Scheduler::now() + backoff, this, packetSave[id]);
      Scheduler::getScheduler().schedule (sendEvent);
    }
    return;
  }

  // from now on: reception of a data packet
  assert (packet->type == PacketType::DATA);
  if (hostNode->getId() == 1) {  // steering node
    if (datacounter[id] == 0) {
      // transmit it at the 1st reception
      PacketPtr packetClone(packet->clone());
      hostNode->enqueueOutgoingPacket (packetClone);
      forwardedDataPackets++;
      packetForwarded[id] = true;
      datacounter[id] = 1;  // discard subsequent packets
    }  // else discard
    return;
  }

  // we suppose that there are nodes on the right of the world in the last commRange/10
  // 10 value is empirical
  assert (packet->dstId == -1);  // code for specific dest is not yet implemented
  if (hostNode->getXPos() > ScenarioParameters::getWorldXSize() -
      ScenarioParameters::getCommunicationRange() / 10) {
    hostNode->dispatchPacketToApplication (packet);
    LogSystem::EventsLogOutput.log (LogSystem::dstReach, Scheduler::now(),
                                    hostNode->getId(), (int)packet->type,
                                    packet->flowId, packet->flowSequenceNumber,
                                    -1, -1, -1);
    Scheduler::getScheduler().endSimulation ();
  }

  if (datacounter[id] == 0) {  // packet id is seen for the first time
    datacounter[id] ++;
    timefirst[id] = Scheduler::now();
    timesecond[id] = -1;
    unscheduled[id] = false;
  } else if (datacounter[id] == 1) { // 2nd packet id received
    assert (timefirst[id] != -1);
    datacounter[id] ++;
    timesecond[id] = Scheduler::now();
    // cout << "test " << hostNode->getId() << " " << Scheduler::now()%ts << " " << timeslot[id] << " " << tp << endl;

    if (hostNode->distance(World::getNode(packet->transmitterId)) > ScenarioParameters::getCommunicationRange() / 3  // not too close
        && ((Scheduler::now() - timefirst[id])%ts < tp
            || (Scheduler::now() - timefirst[id])%ts > ts - tp)) {  // collinearity
      // schedule event e[id] to forward p[id] at now+backoff
      // if this backoff is modified, be sure to update above too (FORGET packet backoff)
      uniform_int_distribution<simulationTime_t> distrib(0, 100000);  // TODO use a computed value, based on the density, instead of 100000
      // term at left of + is the backoff to allow nodes to hear other packets before sending, and term at right of + is the delay to have the same sending time modulo ts as the packet received
      simulationTime_t backoff = packet->size * ts * distrib(*backoffrng) + ts - tp;
      PacketPtr rtf (new Packet (PacketType::RTF, RTF_SIZE, hostNode->getId(), -1, -1, -1, id));
      rtf->disableBackoff = true;
      auto sendEvent = new RayTracingSendEvent (Scheduler::now() + backoff, this, rtf);
      Scheduler::getScheduler().schedule (sendEvent);
      rtfSent[id] = false;
      ctfReceived[id] = false;
      //PacketPtr ppacket(new Packet(_packet));
      PacketPtr packetClone(packet->clone());
      packetSave[id] = packetClone;
    }
  } else {  // 3rd or subsequent packet
    datacounter[id] ++;
    if (unscheduled[id] == false)
      unscheduled[id] = true;  // unschedule, i.e. make it a noop
  }
}

void RayTracingRoutingAgent::processSendEvent (PacketPtr packet) {
  int id = packet->flowSequenceNumber;  // TODO make id unique no matter the src/flow
  assert (id == 0);

  if (packet->type == PacketType::RTF) {
    if (unscheduled[id] == true)
      return;

    if (rtfSent[id] == false) {
      assert (ctfReceived[id] == false);
      // send 1st RTF
      hostNode->enqueueOutgoingPacket (packet);
      //cout << "eugen rtf1 sent by node " << hostNode->getId() << " at " << Scheduler::now() << endl;

      // following RTFs are useful only in shadowing model
      // schedule to send a 2nd RTF if no CTF will be received during that time
      PacketPtr rtf (new Packet (PacketType::RTF, RTF_SIZE, hostNode->getId(), -1, -1, -1, id));
      rtf->disableBackoff = true;
      // +3 to be sure and avoid corner cases
      simulationTime_t backoff = Node::getPulseDuration() * packet->beta * (RTF_SIZE + CTF_SIZE + 3);
      auto sendEvent = new RayTracingSendEvent (Scheduler::now() + backoff, this, rtf);
      Scheduler::getScheduler().schedule (sendEvent);

      // schedule to send a 3rd RTF if no CTF will be received during that time
      PacketPtr rtf2 (new Packet (PacketType::RTF, RTF_SIZE, hostNode->getId(), -1, -1, -1, id));
      rtf2->disableBackoff = true;
      auto sendEvent2 = new RayTracingSendEvent (Scheduler::now() + 2*backoff, this, rtf2);
      Scheduler::getScheduler().schedule (sendEvent2);

      // schedule to send a 4th RTF if no CTF will be received during that time
      PacketPtr rtf3 (new Packet (PacketType::RTF, RTF_SIZE, hostNode->getId(), -1, -1, -1, id));
      rtf3->disableBackoff = true;
      auto sendEvent3 = new RayTracingSendEvent (Scheduler::now() + 3*backoff, this, rtf3);
      Scheduler::getScheduler().schedule (sendEvent3);

      // schedule to send a 5th RTF if no CTF will be received during that time
      PacketPtr rtf4 (new Packet (PacketType::RTF, RTF_SIZE, hostNode->getId(), -1, -1, -1, id));
      rtf3->disableBackoff = true;
      auto sendEvent4 = new RayTracingSendEvent (Scheduler::now() + 4*backoff, this, rtf4);
      Scheduler::getScheduler().schedule (sendEvent4);

      rtfSent[id] = true;
    } else if (ctfReceived[id] == false) {
      // send 2nd, 3rd, 4th or 5th RTF
      hostNode->enqueueOutgoingPacket (packet);
      //cout << "eugen rtf2345 sent by node " << hostNode->getId() << " at " << Scheduler::now() << endl;
    } else {
      // do nothing, because the CTF has been received
    }
  } else if (packet->type == PacketType::DATA) {
    if (unscheduled[id] == true)
      return;

    // forward data packet
    hostNode->enqueueOutgoingPacket (packet);
    //cout << "eugen data sent by node " << hostNode->getId() << " at " << Scheduler::now() << endl;
    forwardedDataPackets++;
    packetForwarded[id] = true;

    // schedule a FORGET packet
    /*static*/ simulationTime_t tp = Node::getPulseDuration();
    /*static*/ simulationTime_t ts = tp * packet->beta;
    PacketPtr forget (new Packet (PacketType::FORGET, FORGET_SIZE, hostNode->getId(), -1, -1, -1, id));
    forget->disableBackoff = true;
    // +3 to be sure and avoid corner cases
    // note that packet->size must be the same as below, i.e. a DATA packet size (which is the case here)
    simulationTime_t backoff = packet->size * ts * 100003 + ts - tp;  // 80 * 100000 fs * 100000 = 0.8 ms !
    auto sendEvent = new RayTracingSendEvent (Scheduler::now() + backoff, this, forget);
    Scheduler::getScheduler().schedule (sendEvent);
  } else if (packet->type == PacketType::FORGET) {
    if (rtfSender[id] == -1) {  // if it has not received any RTF
      hostNode->enqueueOutgoingPacket (packet);
      //cout << "eugen forget sent by node " << hostNode->getId() << " at " << Scheduler::now() << endl;
      packetForwarded[id] = false;
      rtfSender[id] = -1;
      // ensure that this node is not eligible again to send DATA, since it led to die-out
      datacounter[id] += 20;  // pretend to have seen 20 packets
    }  // else ignore this event
  }
}
